@@ -3,13 +3,20 @@
 #  Script to push docs from my development area to SourceForge, where the
 #  update-docs.sh script unpacks them into their final destination.
 
-TARGET=python.sourceforge.net:/home/users/f/fd/fdrake/tmp
+TARGETHOST=www.python.org
+TARGETDIR=/usr/home/fdrake/tmp
+
+TARGET="$TARGETHOST:$TARGETDIR"
 
 ADDRESSES='python-dev@python.org doc-sig@python.org python-list@python.org'
 
-VERSION=`echo '$Revision: 1.13 $' | sed 's/[$]Revision: \(.*\) [$]/\1/'`
+TOOLDIR="`dirname $0`"
+VERSION=`$TOOLDIR/getversioninfo`
+
+# Set $EXTRA to something non-empty if this is a non-trunk version:
 EXTRA=`echo "$VERSION" | sed 's/^[0-9][0-9]*\.[0-9][0-9]*//'`
-if [ "$EXTRA" ] ; then
+
+if echo "$EXTRA" | grep -q '[.]' ; then
     DOCLABEL="maintenance"
     DOCTYPE="maint"
 else
@@ -56,6 +63,10 @@ if [ "$1" ] ; then
     shift
 fi
 
+if [ "$DOCTYPE" = 'maint' ] ; then
+    DOCTYPE='maint23'
+fi
+
 START="`pwd`"
 MYDIR="`dirname $0`"
 cd "$MYDIR"
@@ -65,20 +76,20 @@ cd ..
 
 # now in .../Doc/
 make --no-print-directory bziphtml || exit $?
-RELEASE=`grep '^RELEASE=' Makefile | sed 's|RELEASE=||'`
-PACKAGE="html-$RELEASE.tar.bz2"
+PACKAGE="html-$VERSION.tar.bz2"
 scp "$PACKAGE" tools/update-docs.sh $TARGET/ || exit $?
-ssh python.sourceforge.net tmp/update-docs.sh $DOCTYPE $PACKAGE '&&' rm tmp/update-docs.sh || exit $?
+ssh "$TARGETHOST" tmp/update-docs.sh $DOCTYPE $PACKAGE '&&' rm tmp/update-docs.sh || exit $?
 
 if $ANNOUNCE ; then
     sendmail $ADDRESSES <<EOF
 To: $ADDRESSES
 From: "Fred L. Drake" <fdrake@acm.org>
 Subject: [$DOCLABEL doc updates]
+X-No-Archive: yes
 
 The $DOCLABEL version of the documentation has been updated:
 
-    http://python.sourceforge.net/$DOCTYPE-docs/
+    http://$TARGETHOST/dev/doc/$DOCTYPE/
 
 $EXPLANATION
 EOF
