@@ -2,18 +2,20 @@
 
 Implements the Distutils 'build_scripts' command."""
 
-# created 2000/05/23, Bastian Kleineidam
+# This module should be kept compatible with Python 1.5.2.
 
-__revision__ = "$Id: build_scripts.py,v 1.8 2001/03/02 07:28:03 fdrake Exp $"
+__revision__ = "$Id: build_scripts.py 35629 2004-04-01 03:56:46Z fdrake $"
 
 import sys, os, re
+from stat import ST_MODE
+from distutils import sysconfig
 from distutils.core import Command
 from distutils.dep_util import newer
+from distutils.util import convert_path
+from distutils import log
 
-# check if Python is called on the first line with this expression.
-# This expression will leave lines using /usr/bin/env alone; presumably
-# the script author knew what they were doing...)
-first_line_re = re.compile(r'^#!(?!\s*/usr/bin/env\b).*python(\s+.*)?')
+# check if Python is called on the first line with this expression
+first_line_re = re.compile('^#!.*python[0-9.]*([ \t].*)?$')
 
 class build_scripts (Command):
 
@@ -39,6 +41,8 @@ class build_scripts (Command):
                                    ('force', 'force'))
         self.scripts = self.distribution.scripts
 
+    def get_source_files(self):
+        return self.scripts
 
     def run (self):
         if not self.scripts:
@@ -52,14 +56,16 @@ class build_scripts (Command):
         ie. starts with "\#!" and contains "python"), then adjust the first
         line to refer to the current Python interpreter as we copy.
         """
-        outfiles = []
         self.mkpath(self.build_dir)
+        outfiles = []
         for script in self.scripts:
             adjust = 0
+            script = convert_path(script)
             outfile = os.path.join(self.build_dir, os.path.basename(script))
+            outfiles.append(outfile)
 
             if not self.force and not newer(script, outfile):
-                self.announce("not copying %s (up-to-date)" % script)
+                log.debug("not copying %s (up-to-date)", script)
                 continue
 
             # Always open the file, but ignore failures in dry-run mode --
@@ -80,15 +86,23 @@ class build_scripts (Command):
                 match = first_line_re.match(first_line)
                 if match:
                     adjust = 1
-                    post_interp = match.group(1)
+                    post_interp = match.group(1) or ''
 
             if adjust:
-                self.announce("copying and adjusting %s -> %s" %
-                              (script, self.build_dir))
+                log.info("copying and adjusting %s -> %s", script,
+                         self.build_dir)
                 if not self.dry_run:
                     outf = open(outfile, "w")
-                    outf.write("#!%s%s\n" % 
-                               (os.path.normpath(sys.executable), post_interp))
+                    if not sysconfig.python_build:
+                        outf.write("#!%s%s\n" % 
+                                   (os.path.normpath(sys.executable),
+                                    post_interp))
+                    else:
+                        outf.write("#!%s%s\n" %
+                                   (os.path.join(
+                            sysconfig.get_config_var("BINDIR"),
+                            "python" + sysconfig.get_config_var("EXE")),
+                                    post_interp))
                     outf.writelines(f.readlines())
                     outf.close()
                 if f:
@@ -97,6 +111,18 @@ class build_scripts (Command):
                 f.close()
                 self.copy_file(script, outfile)
 
+        if os.name == 'posix':
+            for file in outfiles:
+                if self.dry_run:
+                    log.info("changing mode of %s", file)
+                else:
+                    oldmode = os.stat(file)[ST_MODE] & 07777
+                    newmode = (oldmode | 0555) & 07777
+                    if newmode != oldmode:
+                        log.info("changing mode of %s from %o to %o",
+                                 file, oldmode, newmode)
+                        os.chmod(file, newmode)
+
     # copy_scripts ()
 
 # class build_scripts
